/*------------------------------------------------------------
- UDPCollectorForEmulator.c
- This program implements the data receiving  behavior of any
- computer via UDP ports.  To send data, the computer must be 
- running the companion file,UDPEmulator.c

- To run the program compile and run executable in command line
  e.g.  <./ExecutableName> <FPGA IP address> <Collector Port>
- As an example if the program has been compiled with the
  executable name 'Collector' and the machine running the 
  UDPEmulator.c code has an IP address of 132.206.124.209 and 
  the port that is opened on the collector is port number 9000, 
  Then run the program as
  
  ./Collector 132.206.124.209 9000


- To view the output file '.txt' use a binary text editor. 
  For a Linux machine the 'bvi editor' (version of vi editor)
  or Emacs is useful.

-  Last modified: Rajat Mukherjee 9/AUG/2006
-------------------------------------------------------------*/
#include <stdio.h> //for printf() and fprintf()
#include <sys/socket.h> //for socket() and bind()
#include <arpa/inet.h> //for sockaddr_in and inet_ntoa()
#include <string.h> //for memset()
#include <stdlib.h> //for atoi()
#include <unistd.h>//for close()
#include <errno.h>//For error messages

//The format for the names of files where data from the FPGA
//will be stored.  This should contain two %d markers.  The
//first corresponds to the board's unique ID number, which is
//taken to be the least significant byte of its IP address.
//The second is the file number, which differentiates between
//subsequent files from this board.
#define FILENAME_FORMAT "data%d_%d.txt"

//The number of files to be created for a given board.  Once
//the maximum number of files has been reached, the first will
//be overwritten.
#define NUM_FILES 25

//The number of data packets (length BYTES_PER_PACKET bytes) to be stored in
//each file.
#define PACKETS_PER_FILE 100

//Defines the size of each data packet to be written (above)
#define BYTES_PER_PACKET 1260

//Defines the total number of files to be received
//For verification purposes only
#define FILES_TO_BE_RECEIVED 3

//Function which transforms a standard-format IP address (ipaddr)
//into a 32-bit computer representation (*output).  The return value
//is a boolean indicating that the string was properly formatted.
char IPaddrToInt(char*ipaddr, unsigned*output);
//Function which transforms a numeric string (str) into an integer
//(*output).  The return value is a boolean indicating that the string
//was properly formatted.
char StrToInt(char*str, unsigned*output);


//Main function, checks the command line arguments for the IP
//address and port number, then sends a UDP frame to that address
//to signal the beginning of data collection.
//Note that UDP is connectionless, so there is no way to guarantee
//that the frame is sent to the correct destination.  Thus, the
//responsibility for inputting the correct IP address and port
//number belongs to the user.
int main(int argc, char*argv[])
{
  //various "nums" describing where data is being written.
  //also the socket descriptor sock.
  int boardnum, packet_counter, filenum, datanum, sock, i,a,position,CharToLongCounter,errno;
  int totalBytesRcvd,recvMsgSize,FpgaAddrLen;
  int counter = 0;//a counter variable for number of files received
  short data[BYTES_PER_PACKET/2], buffer[BYTES_PER_PACKET/2];
  //IP address and port number in machine format.
  unsigned formatted_ipaddr;
  int CollectPort; 
  //data is the data read from the socket; filename is the buffer
  //to hold the name of the current file.
  char filename[50], validentry;//Array for filename and validentry is a flag
  char *ipaddr;//stores the IP address of the FPGA
  //file descriptor for data output file.
  FILE *dataFile;
  //address structure referring to the location of the FPGA on the
  //network.
  struct sockaddr_in FpgaAddr; //For FPGA
  struct sockaddr_in CollectAddr; //For Collector
  
  //Check that the correct number of command line arguments are included,
  //and that these arguments are properly formatted.  Store the computer-
  //formatted IP address and port number to formatted_ipaddr and portnum
  //while checking.
  /*Read the IP address of server and port from command line*/
  if (argc != 3)
    {
      fprintf(stderr,"Usage: %s <FPGA IP Address> <COLLECTOR PORT>\n", argv[0]);
      exit(1);
    }
  ipaddr = argv[1];
  CollectPort = atoi(argv[2]);

  /*Check if the IP address follows the correct format*/
  validentry=IPaddrToInt(ipaddr, &formatted_ipaddr);
  if(!validentry)
    {
      printf("Sorry, %s is not a valid IP address.\n", ipaddr);
      printf("IP addresses consist of four one-byte values separated\n");
      printf("by periods;\nExample: 128.3.100.255\n");
    }
  
  //Check that a valid port number is being requested.
  if(CollectPort>0xffff){
    printf("Port numbers must be between 0 and 65535.\r\n");
    exit(0);
  }
  
  //Create a new socket described by sock for external use (AF_INET)
  //and UDP (SOCK_DGRAM).  The protocol is set to the default value 0.

  printf("About to create UDP socket... \r\n");
  sock=socket(AF_INET, SOCK_DGRAM, 0);
  if(sock<0){
    printf("ERROR: The socket could not be created.\n");
    exit(0);
  }
  printf("UDP Socket is created. \r\n");


  //Set up the address structure to connect to the FPGA:
  //external address,
  //memset(&CollectAddr, 0, sizeof(CollectAddr));
  CollectAddr.sin_family=AF_INET;
  //numeric IP address of the FPGA,
  CollectAddr.sin_addr.s_addr=htonl(INADDR_ANY);
  //Port number.  Note the use of htons(), which enforces the proper byte
  //order on the network.  The FPGA automatically arranges its bytes in
  //the proper order, so this function is not necessary on the server side.
  CollectAddr.sin_port=htons(CollectPort);
  
  
  //Bind Socket
  printf("About to bind UDP socket... \r\n");
  if (bind(sock, (struct sockaddr *) &CollectAddr, sizeof(CollectAddr))<0)
    {
      printf("Binding failed because of Error %d: %s \r\n", errno, strerror(errno));
      exit(0);
    }
  printf("Binding UDP socket complete.Ready to write data... \r\n");
  //Set the board num to the least significant byte of the IP address,
  //and initialize the file counter.
  boardnum=formatted_ipaddr>>24;
  filenum=0;
  FpgaAddrLen = sizeof(FpgaAddr);

  //Run forever. In the actual program this will be 1. For now it is limited.
  while(counter<FILES_TO_BE_RECEIVED)
    //while(1)
    {
      //construct the current filename.
      sprintf(filename, FILENAME_FORMAT, CollectPort, filenum);
      //open the file with filename for writing in binary mode ('wb').
      //if the file already exists, erase its contents.
      dataFile=fopen(filename, "wb");
      if(dataFile==NULL)
	{
	  printf("ERROR: Data file %s could not be opened for writing.\n", filename);
	  exit(0);
	}

      //read PACKETS_PER_FILE packets from the socket, and write their contents
      //to the file.
      for(datanum=0; datanum<PACKETS_PER_FILE; datanum++)
	{
	  totalBytesRcvd = 0;
	  bzero(data,BYTES_PER_PACKET);//clears existing data for TCP
	  bzero(buffer,BYTES_PER_PACKET);
	  position = 0;
	  while(totalBytesRcvd < BYTES_PER_PACKET)
	    {
	      //recvMsgSize = read(sock, buffer, (BYTES_PER_PACKET-totalBytesRcvd));

	      //For UDP
	      recvMsgSize = recvfrom(sock, buffer, BYTES_PER_PACKET-totalBytesRcvd, 0, (struct sockaddr *) &FpgaAddr, &FpgaAddrLen);
	      	      
	      /*Various error options for UDP*/
	      if(recvMsgSize < 0)
		{
		  printf("recvfrom() failed \n");
		  datanum--;
		}
	      else
		{
		  if(recvMsgSize == 0)
		    {
		      //printf("\n No messages to be received");
		      datanum--;
		    }
		  else
		    {
		      a = 0;
		      while(a<(recvMsgSize/2))
			{
			  data[position]=buffer[a];
			  position= position+1;
			  a++;
			}
		      totalBytesRcvd += recvMsgSize;
		      
		    }
		}
	    }
	    
	  //Write the data
	  for(a=0; a<(BYTES_PER_PACKET/2); a++)
	    {
	      if((fprintf(dataFile, "%d ", data[a]))<0)
	      printf("\n Writing character # %d for file %d failed!", a,counter+1);
	    }
	}
      	    
      //print the standard newline character, close the file, and increment
      //the file number.
      if((fprintf(dataFile, "\n"))<0)
	printf("\n Writing eol character for file %d failed!", counter+1);

      if((fclose(dataFile))!=0)
	printf("\n Closing file %d failed!", counter+1);

      filenum=(filenum+1)%NUM_FILES;
      counter++;
      
      //Print confirmation
      printf("Done writing File %d\n\n",counter);

    }
  
}

char IPaddrToInt(char*ipaddr, unsigned*output){
  //Function for converting IP address strings to 32-bit computer
  //integers.  The byte order used by the computer is the reverse of
  //the string (i.e. the least significant byte of the string becomes
  //the most significant byte in the integer representation).
  //INPUTS: ipaddr: pointer to a buffer containing the IP address to
  //                be converted
  //        output: pointer to the memory space where the integer
  //                representation of the IP address will be stored.
  //OUTPUT: a boolean value, 1 if the string was converted successfully
  //        and 0 if it could not be converted because it was not
  //        properly formatted.
  
  int numperiods=0, i, num=-1;
  (*output)=0;
  for(i=0; i<strlen(ipaddr); i++){
    if((ipaddr[i]>='0')&&(ipaddr[i]<='9')){
      if(num<0) num=(ipaddr[i]-'0');
      else num=(10*num)+(ipaddr[i]-'0');
    }
    else if(ipaddr[i]=='.'){
      if((num<0)||(num>255)) return 0;
      (*output)+=num<<(numperiods*8);
      numperiods++;
      if(numperiods>3) return 0;
      num=-1;
    }
    else return 0;
  }
  (*output)+=num<<24;
  return 1;
}

char StrToInt(char*input, unsigned*output){
  //Function for converting a numeric string to a positive integer.
  //The string also error-checks the input string for proper formatting.
  //INPUTS: input: a pointer to a buffer containing the string to be
  //               converted
  //        output: a pointer to the space in memory where the integer
  //               representation of the string will be placed.
  //OUTPUT: a boolean value, 1 if the conversion was successful and 0
  //        if the conversion failed because the string contained
  //        non-numeric characters.

  int i;
  for(i=0; i<strlen(input); i++){
    if((input[i]>='0')&&(input[i]<='9'))
      (*output)=((*output)*10)+(input[i]-'0');
    else return 0;
  }
  return 1;
}
