/*--------------------------------------------------------------------
- UDPEmulator.c
- This program implements the data sending behavior of any terminal
- computer via UDP ports.  To accept data, the computer must be 
- running the binary compiled from the companion file, UDPCollector.c.

- The WhiteNoiseGenerator() and the UniformDeviateRandomNumber() 
  functions are for returning a gaussian random variable and are 
  versions of the gasdev() and ran1() functions respectively from the 
  source "Numerical Recipes in C: The Art of Scientific Computing", 
  Cambridge University Press, 1992. This is online at the link 
  http://www.library.cornell.edu/nr/cbookcpdf.html

- To compile sometimes the gcc Linux compiler doesn't recognize the 
  functions log and sqrt. Hence use the option -lm while executing.
  As an example if the program has been compiled with the
  executable name 'Emulator' and is sending packets to the Collector 
  that has an an IP address of 132.206.124.201 and 
  the port that is opened on the collector is port number 9000, 
  Then run the program as
  
  ./Emulator 132.206.124.201 9000 -lm

-  Last modified: Rajat Mukherjee 9/AUG/2006
--------------------------------------------------------------------*/

#include <stdio.h> //for printf() and fprintf()
#include <sys/socket.h> //for socket() and bind()
#include <arpa/inet.h> //for sockaddr_in and inet_addr().inet_addr() is a built in function
//to convert IP address in dotted quad format to 32 bit binary equivalent.
#include <string.h> //for memset()
#include <stdlib.h>//for atoi()
#include <unistd.h>//for close()
#include <time.h> //for sleep mode
#include <math.h>   //for random number generation

//Define the number of samples per each packet to be 12
#define SAMPLES_PER_PACKET 12

//Define the number of bolometers per FPGA to be 48
#define BOLOS_PER_FPGA 48

//Defines the size of each data packet
#define BYTES_PER_PACKET SAMPLES_PER_PACKET*(2*BOLOS_PER_FPGA+8)+12

//Defines the time to sleep for each data frame (in microseconds)
#define SLEEPTIME 50000

//Defines the total number of packets to be sent;only for verification
#define PACKETS_TO_BE_SENT 300

//Defines whether to use random noise in data generation or not
//Set to 1 if noise is desired or set to 0 otherwise
#define ADD_WHITE_NOISE 1

//Defines the parameters of the white noise. Mean should be an 
//integer but the standard deviation can be a float
#define MEAN 0
#define STANDARD_DEVIATION 4


//Defines some parameters necessary for UniformDeviateRandomNumber()
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

int main(int argc, char *argv[])
{

  //Declare functions to be used
  long CalculateChecksum (short *array);
  short WhiteNoiseGenerator(long *idum);

  int sock;//initializes socket
  int indicator;
  int bolocounter;//This is for counting bolometers
  long samplecounter;//this is for counting samples
  long packetcounter,i;//initializes a few counters
  long checksum;//variable for storing checksum
  long *randomseed;//Used for generating a seed for white noise
  unsigned short CollectPort; //specifies the port at which Collector opens socket to receive data

  char *CollectIP;//stores IP address of collector
  short data[BYTES_PER_PACKET/2];//stores the data frame
  short FIFO[PACKETS_TO_BE_SENT][SAMPLES_PER_PACKET][BOLOS_PER_FPGA];//Mimics a FIFO

  //Define timestamp variables
  short MajorPeriod = 0;//Least Significant 14 bits count
  long TickFromTimeServer = 0;//Least significant 18 bits count		    
  long ClientTick = 0;//FPGA time

  struct sockaddr_in CollectAddr; //Identifies Collector Port
  
  /*Check for correct number of command line arguments*/
  if (argc != 3)
    {
      fprintf(stderr,"Enter variables: %s <COLLECTOR IP> <COLLECTOR PORT> \n", argv[0]);
      exit(1);
    }

  /*Command Line Argument 1 is IP address of collector*/
  CollectIP = argv[1];

  /*Command Line Argument 2 is Port number of collector*/
  CollectPort = atoi(argv[2]);

  //Construct the packet header (the same for all packets for a given FPGA)
  data[0]=0x5555;//part of cookie
  data[1]=0x5555;//part of cookie
  data[2]=(0x8000|CollectPort);
  //MSB is the number of bits per sample and last 15 bits are the bolo ID set to the 
  //particular port being connected to
  data[3]= 0x300c;//First 8 bits give no. of bolometers (48) 
                  //and last 8 bits give number of time sample (12)

  //Create FIFO entries for SAMPLES_PER_PACKET number of samples depending
  //on value of ADD_WHITE_NOISE
  if (ADD_WHITE_NOISE == 0)
    {
      for(packetcounter=0;packetcounter<PACKETS_TO_BE_SENT;packetcounter++)
	{
	  for(samplecounter=0;samplecounter<SAMPLES_PER_PACKET;samplecounter++)
	    {
	      for(bolocounter=0;bolocounter<BOLOS_PER_FPGA;bolocounter++)
		{
		  //Create data in FIFO as straight count
		  FIFO[packetcounter][samplecounter][bolocounter]= 
		    (packetcounter*SAMPLES_PER_PACKET*BOLOS_PER_FPGA) 
		    + samplecounter*(BOLOS_PER_FPGA) + bolocounter+1;
		}
	    }
	}
    }
  else
    {
      for(bolocounter=0;bolocounter<BOLOS_PER_FPGA;bolocounter++)
      	{
	  for(packetcounter=0;packetcounter<PACKETS_TO_BE_SENT;packetcounter++)
	    {
	      for(samplecounter=0;samplecounter<SAMPLES_PER_PACKET;samplecounter++)
		{
		  //Create new random seed for each channel (i.e. bolometer)
		  if((packetcounter == 0) && (samplecounter == 0))
		    {
		      static long store = 0;
		      while(store >= 0)
			{
			  store = (long)((-1)*rand());
			}
		      randomseed = &store;
		    }
		  //Invoke Gaussian generator
		  FIFO[packetcounter][samplecounter][bolocounter]= WhiteNoiseGenerator(randomseed);
		}
	    }
	}
    }

  /*Creates a UDP socket*/
  if ((sock = socket(PF_INET,SOCK_DGRAM,IPPROTO_UDP))<0)
    {
      printf("Error creating socket\n");
      exit(0);
    }

  memset(&CollectAddr, 0, sizeof(CollectAddr));//memset() sets all unused structure variables to zero 
  CollectAddr.sin_family = AF_INET;//Internet address family
  CollectAddr.sin_addr.s_addr = inet_addr(CollectIP);//Collector IP address
  CollectAddr.sin_port = htons(CollectPort);//Collector port. htons() ensures byte ordering

  /*Send data. Each time the WHILE loop repeats it is a new packet in the collector*/
  packetcounter = 1;
  while(packetcounter<=PACKETS_TO_BE_SENT)
  //while(1)
  {
    indicator = 4;
      //Read 12 samples and add time stamps
      
      for(samplecounter=1;samplecounter<=SAMPLES_PER_PACKET;samplecounter++)
	{
	  //Create timestamp
	  //Stores the first 14 bits as major period
	  data[indicator]=MajorPeriod<<2;
	  //last 2 bits of current data[indicator] are the 2 MSBs of tick from time server
	  data[indicator]= (short)(data[indicator]|((TickFromTimeServer>>16)&0x0003));
	  
	  //Next data[indicator] stores the least significant 16 bits of TickFromTimeServer
	  data[++indicator]= (short)(TickFromTimeServer & 0xffff);
	  
	  //Client Tick is incremented by 523 for each new sample
	  ClientTick = ClientTick+523;

	  //Most significant 16 bits of CLientTick stored in the next data[indicator]
	  data[++indicator]= (short)((ClientTick >> 16)&0xffff);
	  //Least significant 16 bits of CLientTick stored in the next data[indicator]
	  data[++indicator]= (short)(ClientTick & 0xffff);
	  indicator++;
	  //Timestamp created.Ready to append data
	  for(bolocounter=0;bolocounter<BOLOS_PER_FPGA;bolocounter++)
	    {
	      //Read data from FIFO
	      data[indicator]= FIFO[packetcounter-1][samplecounter-1][bolocounter];
	      indicator++;
	    }
	}
      //Calculate checksum
      checksum = CalculateChecksum (data);

      //Append checksum to the end of the data packet. Checksum is 32 bits so its split in half
      data[628]=(checksum >> 16);//Most significant 16 bits of checksum
      data[629]=(checksum & 0xffff);//Least significant 16 bits of checksum
      
      //Send data
      if(sendto(sock,data,BYTES_PER_PACKET,0,(struct sockaddr *) &CollectAddr, sizeof(CollectAddr))!=BYTES_PER_PACKET)
	printf("\n sendto() sent a different number of bytes than expected");
      
      
      usleep (SLEEPTIME);//If a packet has been sent it sleeps for 50ms(~10 samples)
      printf("Done sending Packet %d \n", packetcounter);

      //MajorPeriod increments every 6 hours or 343,800 packets
      if((packetcounter%343800)==0)
	{
	  MajorPeriod++;
	}

      //Tick from Time Server increments every 250ms or 48 samples (4 packets)
      //Also reset ClientTick every time sync signal is received
      if((packetcounter%4)==0)
	{
	  TickFromTimeServer++;
	  ClientTick = 0;
	}

      packetcounter++;
  }

  //Close socket
  close(sock);
  exit(0);
}






//Sub-routine for calculating checksum for a given packet
long CalculateChecksum (short *array)
{
  int loop;
  long checksum = 0,temp;
  //Run a loop 
  for (loop = 0;loop<((BYTES_PER_PACKET-2)/2);loop++)
    {
      //Store the first 16-bit word as the MSB 
      temp = (long) array[2*loop]<<16;
      //Store the next 16-bit word as the LSB
      temp = temp |((long) array[2*loop+1]);
      //Add 'temp' to the current value of checksum
      checksum += temp;
    }
  //The actual checksum is calculated by summing all 32-bit words and then subtracting 
  //from 0
  checksum = 0 - checksum;
  return checksum;
}



//Returns a normally distributed variate with defined mean and variance, using 
//UniformDeviateRandomNumber(idum) as the source for uniform deviates
short WhiteNoiseGenerator(long *idum)
{
  float UniformDeviateRandomNumber(long *idum);
  short RoundingAlgorithm(double number);

  static int iset = 0;//stays unchanged over different routine calls
  static short number2;
  short number1;
  double fac,rsq,v1,v2,numbefround1,numbefround2;
  if(*idum <0)
    iset = 0;//Re-initialize
  
  //We don't have an extra deviate handy, so pick two numbers in the square 
  //extending from -1 to +1 in each direction, see if they are in the unit circle,
  //and if they are not then try again
  if(iset==0)
    {
      do
	{
	  v1=(double)(2.0*UniformDeviateRandomNumber(idum)-1.0);
	  v2=(double)(2.0*UniformDeviateRandomNumber(idum)-1.0);
	  rsq=(double)(v1*v1+v2*v2);
	}
      while (rsq >= 1.0 || rsq == 0.0);
      fac=(double)(sqrt(-2.0*log(rsq)/rsq));

      //Now make the Box-Muller transformation to get two normal deviates.
      //Return one and save the other one for next time.
      numbefround1 = ((v1*fac*STANDARD_DEVIATION)+MEAN);//first deviate
      number1 = RoundingAlgorithm(numbefround1);//Round answer and return

      iset=1;//Set flag for next time

      numbefround2 = (( v2*fac*STANDARD_DEVIATION)+MEAN);//second deviate
      number2 = RoundingAlgorithm(numbefround2);//Round answer and return
      return number1;
    }
  else //We have an extra deviate handy, so unset the flag, and return it.
    {
      iset = 0;
      return number2;
    }
}


//"Minimal" random number generator of Park and Miller with Bays-Durham shuffle and added
//safeguards. Rerturns a uniform deviate between 0.0 and 1.0 (excluding end-points). Call
//with idum a negative integer to initialize; thereafter, do not alter idum between successive
//deviates in a sequence. RNMX should approximate the largest floating value that is < 1
float UniformDeviateRandomNumber(long *idum)
{
  int j;
  long k;
  static long iy = 0;
  static long iv[NTAB];
  float temp;
  if(*idum <= 0 || !iy)
    {
      //Initialize
      if(-(*idum)<1)
	*idum = 1;//Be sure to prevent idum = 0;
      else
	*idum = -(*idum);
      for(j = NTAB+7;j>=0;j--)
	{
	  //Load the shuffle table (after 8 warm-ups)
	  k=(*idum)/IQ;
	  *idum= IA*(*idum-k*IQ)-IR*k;
	  if(*idum<0)
	    *idum += IM;
	  if(j<NTAB)
	    iv[j] = *idum;
	}
      iy=iv[0];
    }
  k = (*idum)/IQ;//Start here when not initializing
  //Compute idum=(IA*idum)%IM without overflows by Schrage's method
  *idum=IA*(*idum-k*IQ)-IR*k;
  if(*idum < 0)
    *idum += IM;
  j=iy/NDIV;//Will be in the range 0 ... NTAB-1
  iy=iv[j];//Output previously stored value and refill the shuffle table
  iv[j]=*idum;
  if((temp=AM*iy) > RNMX)
    return RNMX;//Because users do not expect endpoint values
  else
    return temp;
}

//Rounding Algorithm for adjusting bins in Gaussian generator
short RoundingAlgorithm (double x)
{
  short rounded;
  rounded = (short)x;
  
  if(x < 0)
    {
      if((rounded - x)>=0.5)
	rounded -= 1;
    }
  if(x > 0)
    {
      if((x - rounded)>=0.5)
	rounded += 1;
    }
  if(x == 0)
    rounded = 0;

  return rounded;
}
    

   
    
     
    


